"{ Package: 'stx:libscm/mercurial' }"

ByteArray variableByteSubclass:#HGNodeId
	instanceVariableNames:'revno'
	classVariableNames:'NullId'
	poolDictionaries:''
	category:'SCM-Mercurial-Core'
!


!HGNodeId class methodsFor:'instance creation'!

fromBytes: aByteArrayOrString

    | sz |

    sz := aByteArrayOrString size.
    (sz ~~ 20 and:[sz ~~ 6]) ifTrue:[
        self error:'Node ID has either 20 or 6 bytes (short form)'.
        ^nil.
    ].
    ^(self new: sz) replaceBytesFrom: 1 to: sz with: aByteArrayOrString startingAt: 1

    "Created: / 25-09-2012 / 21:00:06 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 13-11-2012 / 16:47:40 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

fromString: aString
    ^self readFrom: aString readStream

    "Created: / 10-09-2012 / 10:49:43 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 13-11-2012 / 16:49:39 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

new
    ^self new: 20

    "Created: / 10-09-2012 / 10:42:39 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 19-10-2012 / 15:51:45 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

new: size
    (size ~~ 20 and:[size ~~ 6]) ifTrue:[
        self error: 'Size of HGNodeId must be either 20 bytes or 6 bytes (short form) '.
        ^nil.
    ].
    ^super new: size

    "Created: / 10-09-2012 / 10:44:00 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 13-11-2012 / 16:48:38 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

readFrom: aStringOrStream 
    ^self readFrom: aStringOrStream onError:[:msg|self error:msg].

    "Created: / 13-11-2012 / 16:56:58 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

readFrom: aStringOrStream onError: aBlock
    "Parses node id from stream and returns it. Support both,
     short and full node ids"

    | stream c c2 sign revno hash short |

    stream := aStringOrStream readStream.
    stream peek == $- ifTrue:[        
        stream next.
        sign := -1.
    ] ifFalse:[
        sign := 1.
    ].

    "/ Read revno...
    revno := 0.
    [ (c := stream peek) == $: ] whileFalse:[
        c isDigit ifFalse:[
            self error:'Digit ([0-9]) expected but ', c , 'found'.
        ].
        revno := (revno * 10) + c digitValue.
        stream next.
    ].
    stream next. "/eat :
    revno := revno * sign.

    "/ Read hash
    hash := ByteArray new: 20.
    short := true.
    1 to: 6 do:[:i|
        stream atEnd ifTrue:[
            aBlock valueWithOptionalArgument:'Unexpected end of stream, hex digit expected'.
        ].
        c := stream peek.
        c isHexDigit ifFalse:[
            aBlock valueWithOptionalArgument:'Hex digit ([0-9a-z]) expected but ', c , ' found'.
        ].
        stream next.
        stream atEnd ifTrue:[
            aBlock valueWithOptionalArgument:'Unexpected end of stream, hex digit expected'.
        ].
        c2 := stream peek.
        c2 isHexDigit ifFalse:[
            aBlock valueWithOptionalArgument:'Hex digit ([0-9a-z]) expected but ', c , ' found'.
        ].
        hash at:i put: (c digitValue << 4) | c2 digitValue.
        stream next.
    ].
    (stream atEnd not and:[stream peek isHexDigit]) ifTrue:[
        "/OK, full 40-char node id
        short := false.
        7 to: 20 do:[:i|
            stream atEnd ifTrue:[
                aBlock valueWithOptionalArgument:'Unexpected end of stream, hex digit expected'.
            ].
                        c := stream peek.
            c isHexDigit ifFalse:[
                aBlock valueWithOptionalArgument:'Hex digit ([0-9a-z]) expected but ', c , ' found'.
            ].
            stream next.
            stream atEnd ifTrue:[
                aBlock valueWithOptionalArgument:'Unexpected end of stream, hex digit expected'.
            ].
            c2 := stream peek.
            c2 isHexDigit ifFalse:[
                aBlock valueWithOptionalArgument:'Hex digit ([0-9a-z]) expected but ', c , ' found'.
            ].
            hash at:i put: (c digitValue << 4) + c2 digitValue.
            stream next.
        ].
    ].
    (revno == -1) ifTrue:[
        hash = #[0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0] ifTrue:[
            ^self null
        ].
        hash = #[0 0 0 0 0 0] ifTrue:[
            ^self null
        ].
    ].
    ^short ifTrue:[
        (HGNodeId new: 6)
            revno: revno;
            replaceBytesFrom: 1 to: 6 with: hash startingAt: 1;
            yourself
    ] ifFalse:[
        (HGNodeId fromBytes: hash) revno: revno.
    ]

    "
    HGNodeId fromString:'4:6f88e1f44d9eb86e0b56ca15e30e5d786acd83c7'

    Bad ones:

    HGNodeId fromString:'4:6f88e1f44d9eb86e0b56ca15e30e5d786acd' 
    HGNodeId fromString:'4:6f88Z1f44d9eb86e0b56ca15e30e5d786acd83c7' 
    "

    "Created: / 13-11-2012 / 16:49:19 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!HGNodeId class methodsFor:'accessing'!

null
    NullId isNil ifTrue:[
        NullId := self new.
        NullId revno: -1
    ].
    ^NullId

    "
        HGNodeId null
    "

    "Created: / 19-10-2012 / 15:51:25 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (comment): / 13-11-2012 / 09:54:14 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!HGNodeId methodsFor:'accessing'!

revno
    ^ revno ? -2

    "Created: / 13-11-2012 / 09:52:02 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

revno:anInteger
    revno := anInteger.
! !

!HGNodeId methodsFor:'comparing'!

= anotherId

    self class == anotherId class ifFalse:[ ^ false].
    self size == anotherId size ifTrue:[
        ^super = anotherId
    ].
    "One of them must be short, another long"
    1 to: 6 do:[:i|
        (self at:i) ~~ (anotherId at:i) ifTrue:[ ^ false ].
    ].
    ^true

    "Created: / 13-11-2012 / 17:37:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!HGNodeId methodsFor:'printing & storing'!

displayOn:aStream
    ^self printOn: aStream

    "Created: / 13-11-2012 / 09:55:00 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

printOn:aStream
    "append a printed representation if the receiver to the argument, aStream"

    | rn |

    rn := self revno.
    rn ~~ -2 ifTrue:[
        rn printOn: aStream.
        aStream nextPut: $:.
    ].

    self hexPrintOn:aStream.

    "Modified: / 13-11-2012 / 18:01:45 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!HGNodeId class methodsFor:'documentation'!

version_SVN
    ^ '$Id::                                                                                                                        $'
! !
